#ifndef STARSProgram_h
#define STARSProgram_h

#include <list>
#include <map>
#include <set>
#include <string>

#include <cstdint>
#include <cstdio>

#include "interfaces/STARSTypes.h"
#include "interfaces/SMPDBInterface.h"
#include "interfaces/abstract/STARSInstructionID.h"

// #include "base/SMPProgram.h"
class SMPProgram;

// Optimizing annotation categories
#define LAST_OPT_CATEGORY  10

class STARS_Program_t
{
	public:
		virtual ~STARS_Program_t() { };
		// Data initialization
		virtual void InitData(void);
		virtual void DetermineRootFileName(void) = 0;
		virtual bool OpenFiles(void);
		virtual void CloseFiles(void);
		virtual bool ReopenAnnotFile(const char *mode);
		void ZST_InitPolicies(void);

		// Get (accessor) methods
		std::size_t GetSTARS_ISA_Bitwidth(void) const { return STARS_ISA_Bitwidth; };
		std::size_t GetSTARS_ISA_Bytewidth(void) const { return STARS_ISA_Bytewidth; };
		char GetSTARS_ISA_dtyp(void) const { return STARS_ISA_dtyp; };
		int GetSTARS_MD_LAST_SAVED_REG_NUM(void) const { return STARS_MD_LAST_SAVED_REG_NUM; };
		int GetOptCategory(uint16_t IDAOpcodeEnum) const { return OptCategory[IDAOpcodeEnum]; };
		STARS_sval_t GetStackAlteration(uint16_t IDAOpcodeEnum) const { return StackAlteration[IDAOpcodeEnum]; };
		int GetOptCount(std::size_t CategoryIndex) const { return OptCount[CategoryIndex]; };
		int GetAnnotationCount(std::size_t CategoryIndex) const { return AnnotationCount[CategoryIndex]; };
		unsigned long GetDataReferentID(void) const { return DataReferentID; }; // Unique ID for data annotations
		unsigned int GetShadowID(void) const { return ShadowID; }; // shadowing defenses index for shadowed values
		uint16_t GetCurrentFileNum(void) const { return CurrentFileNumber; };
		std::string GetRootFileName(void) const { return RootFileName; }; // e.g. "foo.exe" when analyzing file "/usr/foo.exe"
		std::string GetAdaPackageName(void) const { return PackageName; }; // e.g. "foo-bar" when analyzing "/usr/foo_bar.exe"
		std::string GetAnnotFileName(void) const { return AnnotFileName; }; // e.g. "foo.exe.annot"
		std::string GetInfoAnnotFileName(void) const { return InfoAnnotFileName; }; // e.g. "foo.exe.infoannot"
		FILE *GetAnnotFile(void) const { return STARS_AnnotFile; };
		FILE *GetInfoAnnotFile(void) const { return STARS_InfoAnnotFile; };
		FILE *GetAlarmFile(void) const { return ZST_AlarmFile; };
		FILE *GetXrefsFile(void) const { return STARS_XrefsFile; };
		FILE *GetCallReturnFile(void) const { return STARS_CallReturnFile; };
		FILE *GetSPARKHeaderFile(void) const { return ZST_SPARKHeaderFile; };
		FILE *GetSPARKSourceFile(void) const { return ZST_SPARKSourceFile; };
		std::list<uint16_t>::iterator GetFirstCallerSavedReg(void) { return STARS_MDCallerSavedRegs.begin(); };
		std::list<uint16_t>::iterator GetLastCallerSavedReg(void) { return STARS_MDCallerSavedRegs.end(); };
		std::list<uint16_t>::iterator GetFirstArgumentReg(void) { return STARS_MDArgumentRegs.begin(); };
		std::list<uint16_t>::iterator GetLastArgumentReg(void) { return STARS_MDArgumentRegs.end(); };
		std::list<uint16_t>::iterator GetFirstCalleeSavedReg(void) { return STARS_MDCalleeSavedRegs.begin(); };
		std::list<uint16_t>::iterator GetLastCalleeSavedReg(void) { return STARS_MDCalleeSavedRegs.end(); };
		ZST_SysCallType GetCallTypeFromFuncName(std::string SysCallName) const;
		ZST_Policy GetPolicyFromCallType(ZST_SysCallType CallType) const;
		SMP_bounds_t GetFuncBounds(std::size_t FuncIndex) const { return FuncBounds.at(FuncIndex); };
		std::size_t GetFuncBoundsSize(void) const { return FuncBounds.size(); };

		// Set (mutator) methods
		void Set32BitBinary(void) { SetBitwidth32(); }; // Set internal state to handle a 32-bit binary
		void Set64BitBinary(void) { SetBitwidth64(); }; // Set internal state to handle a 64-bit binary
		void SetIDAProDriverFlag(void) { IDAProDriver = true; };
		void SetRootFileName(std::string NewName) { RootFileName = NewName; AnnotFileName = NewName + ".annot"; InfoAnnotFileName = NewName + ".infoannot"; };
		void IncrementOptCount(std::size_t OptCategory) { ++OptCount[OptCategory]; }; // increment optimizing annotation count
		void IncrementAnnotationCount(std::size_t OptCategory) { ++AnnotationCount[OptCategory]; }; // increment total annotation count
		void IncrementDataReferentID(void) { ++DataReferentID; }; // increment unique data annotations referent ID
		void IncrementShadowID(void) { ++ShadowID; }; // increment index for items covered by shadowing defenses

		// Query methods
		bool ShouldSTARSPerformReducedAnalysis(void) const { return STARS_PerformReducedAnalysis; }; // Is program too big for full analysis?
		bool IsIDAProDriverMode(void) const { return IDAProDriver; }; 
		bool IsLocationWhitelisted(ZST_SysCallType CallType, std::string LocationName) const;
		bool IsLocationBlacklisted(ZST_SysCallType CallType, std::string LocationName) const;
		// Given a called function name, does it produce only benign numeric errors when
		//  its returned values are used in arithmetic? (i.e. it is a trusted input)
		bool IsNumericSafeSystemCall(std::string CallName) const;
		virtual bool AreInstIDsInSameFunction(const STARS_ea_t InstID1, const STARS_ea_t InstID2) const;
		virtual bool IsCodeAddressTaken(const STARS_ea_t InstAddr) const { return false; };

		// Printing methods

		// Utility functions to print code xrefs to STARS_XrefsFile
		bool PrintCodeToCodeXref(STARS_ea_t FromAddr, STARS_ea_t ToAddr, std::size_t InstrSize, IBType TypeCode);
		bool PrintReturnInstXref(STARS_ea_t RetInstAddr, STARS_ea_t TargetAddr, std::size_t InstrSize, bool TailCallTargetAddr);
		void PrintDataToCodeXref(STARS_ea_t FromDataAddr, STARS_ea_t ToCodeAddr, std::size_t InstrSize);
		void PrintUnknownCodeXref(STARS_ea_t ToAddr, std::size_t InstrSize, IBType TypeCode);
		void PrintCodeToCodeXrefComplete(STARS_ea_t FromAddr, std::size_t InstrSize, std::size_t IBTCount, IBType TypeCode);
		virtual void PrintAllCodeToCodeXrefs(STARS_ea_t InstAddr, std::size_t InstSize, bool CallFlag) = 0;

		// Analysis methods
		virtual bool IsImmedCodeAddress(STARS_ea_t ImmedValue) { return false; }; // Is ImmedValue a code address?
		virtual void ReportTotalCodeSize(unsigned long long TotalCodeSize) = 0; // Set flags, take actions based on code size.
		virtual void InitStaticDataTable(SMPProgram *CurrProg) = 0; // Process global static data sections
		virtual void GetBlockSuccessorTargets(bool CallFlag, STARS_InstructionID_t LastBlockInst, std::size_t InstSize, std::list<STARS_InstructionID_t> &SuccList) = 0; // Get successors for LastBlockInst, omitting call targets
		virtual void FindCodeAddressesTaken(void) = 0;

	protected:
		// We maintain a list of the caller-saved regs for the current binary's ABI.
		//  This differs from 32-bit to 64-bit x86 binaries, as well as across other ISAs.
		//  Ditto for callee-saved regs.
		std::list<uint16_t> STARS_MDCallerSavedRegs;
		std::list<uint16_t> STARS_MDCalleeSavedRegs;

		// We maintain a list of the argument-passing regs for the current binary's ABI.
		//  This differs from 32-bit to 64-bit x86 binaries, as well as across other ISAs.
		//  The list is in order of argument position number. For x86-64, this means EDI,
		//  ESI, EDX, ECX, R8, R9.
		std::list<uint16_t> STARS_MDArgumentRegs;

		// Function address bounds for each func, by IDA Pro index.
		std::vector<SMP_bounds_t> FuncBounds;

		// Mutators
		void SetBitwidth32(void) {
			STARS_ISA_Bytewidth = 4; STARS_ISA_Bitwidth = 32; STARS_ISA_dtyp = STARS_dt_dword;
			STARS_MD_LAST_SAVED_REG_NUM = STARS_x86_R_di;
		};
		void SetBitwidth64(void) {
			STARS_ISA_Bytewidth = 8; STARS_ISA_Bitwidth = 64; STARS_ISA_dtyp = STARS_dt_qword;
			STARS_MD_LAST_SAVED_REG_NUM = STARS_x86_R_r15;
		};
		void SetTotalCodeSize(unsigned long long TotalSize) { STARS_TotalCodeSize = TotalSize; };
		void SetReducedProcessingFlag(bool FlagValue) { STARS_PerformReducedAnalysis = FlagValue; };
		void IncrementCurrentFileNum(void) { ++CurrentFileNumber; };

		// Analysis methods
		void MDInitializeCallerSavedRegs(void);
		void MDInitializeArgumentRegs(void);
		void MDInitializeCalleeSavedRegs(void);

		// Convert a call type string from the policy file, such as "FILECALLS", to the
		//  corresponding ZST_SysCallType, such as ZST_FILE_CALL.
		ZST_SysCallType ConvertStringToCallType(char *Str2);
		// Convert a policy string from the policy file, such as "DISALLOW", to
		//  the corresponding ZST_Policy value, such as ZST_DISALLOW.
		ZST_Policy ConvertStringToPolicy(char *Str3);

	private:
		// Data members.
		std::size_t STARS_ISA_Bitwidth;
		std::size_t STARS_ISA_Bytewidth;
		char STARS_ISA_dtyp;
		bool IDAProDriver;  // program is compiled in IDA Pro mode
		int STARS_MD_LAST_SAVED_REG_NUM;

		// Unique data referent number to use in data annotations.
		unsigned long DataReferentID;

		unsigned int ShadowID; // Unique index into shadowed values (e.g. func ptrs and critical data)

		bool STARS_PerformReducedAnalysis; // true means program is too big for full analysis

		// Total code size for the program in bytes.
		unsigned long long STARS_TotalCodeSize;

		// File number, used as index into vectors of file pointers, maps of function pointers, etc.
		uint16_t CurrentFileNumber;

		// Filename (not including path) of executable being analyzed.
		std::string RootFileName;

		// Filename modified for Ada requirements.
		std::string PackageName;

		// Annotation file names.
		std::string AnnotFileName, InfoAnnotFileName;

		// Files for annotations.
		FILE *STARS_AnnotFile;
		FILE *STARS_InfoAnnotFile;

		// File to print security alert messages to, e.g. foo.exe.alarms.
		FILE *ZST_AlarmFile;

		// File for code xref targets (helps ILR, makes IRDB more complete)
		FILE *STARS_XrefsFile;

		// File to provide details on fast returns, safe and unsafe return-address functions, etc.
		FILE *STARS_CallReturnFile;

		// Files for SPARK Ada translation of program RTLs
		FILE *ZST_SPARKSourceFile;
		FILE *ZST_SPARKHeaderFile;

		// Keep statistics on how many instructions we saw in each optimization
		//  category, and how many optimizing annotations were emitted for
		//  each category.
		int AnnotationCount[LAST_OPT_CATEGORY + 1];
		int OptCount[LAST_OPT_CATEGORY + 1];

		// Define optimization categories for instructions.
		int OptCategory[STARS_NN_last + 1];

		// Record which opcodes change the stack pointer, and by how many
		//  bytes up (reduction in stack size for stacks that grow downward)
		//  or down (increase in stack size for stacks that grow downward).
		STARS_sval_t StackAlteration[STARS_NN_last + 1];

		// Map library function names to their system call type.
		std::map<std::string, ZST_SysCallType> ZST_FuncTypeMap;

		// Map system call types to their Zephyr Security Toolkit security policy.
		std::map<ZST_SysCallType, ZST_Policy> ZST_TypePolicyMap;

		// Set of whitelisted file locations.
		std::set<std::string> ZST_FileLocWhitelist;

		// Set of whitelisted network locations.
		std::set<std::string> ZST_NetworkLocWhitelist;

		// Set of blacklisted file locations.
		std::set<std::string> ZST_FileLocBlacklist;

		// Set of blacklisted network locations.
		std::set<std::string> ZST_NetworkLocBlacklist;

		// Set of system call names whose returned values should be trusted to have only benign numeric errors.
		std::set<std::string> ZST_SystemCallNumericWhitelist;

		// Methods

		// Initialize the OptCategory[] array.
		void InitOptCategory(void);
		// Initialize the StackAlteration[] array.
		void InitStackAlteration(void);
		void InitLibFuncFGInfoMaps(void);
		void InitDFACategory(void);
		void InitSMPDefsFlags(void);
		void InitSMPUsesFlags(void);
		void InitTypeCategory(void);

		void PrintTypeCode(IBType TypeCode);
		virtual void Dump() { } 

};

#endif
